import json
import math
import os
import random
import subprocess
import sys
import time
from pathlib import Path
from typing import NamedTuple, Optional

import openai
import pytest
import requests
from transformers import AutoTokenizer
from vllm import SamplingParams
from vllm.entrypoints.openai.cli_args import make_arg_parser
from vllm.utils import FlexibleArgumentParser, get_open_port
from vllm.v1.request import Request

EmbeddingWarmupShapes = list[tuple[int, int]]
DecodeWarmupShapes = list[tuple[int, int, int]]


def patch_environment(use_cb: bool, warmup_shapes: DecodeWarmupShapes | None,
                      backend: str, monkeypatch):
    # Setup the environment correctly for the LLM

    # ---- For static batching ----
    if warmup_shapes:
        assert not use_cb, ("Warmup shapes through environment variables have "
                            "been deprecated in continuous batching")

        patch_warmup_shapes(warmup_shapes, monkeypatch)

    # --------------
    monkeypatch.setenv("VLLM_SPYRE_USE_CB", "1" if use_cb else "0")
    monkeypatch.setenv("VLLM_SPYRE_DYNAMO_BACKEND", backend)


def patch_warmup_shapes(warmup_shapes: DecodeWarmupShapes
                        | EmbeddingWarmupShapes, monkeypatch):
    warmup_prompt_length = [t[0] for t in warmup_shapes]
    warmup_batch_size = [t[-1] for t in warmup_shapes]

    monkeypatch.setenv('VLLM_SPYRE_WARMUP_PROMPT_LENS',
                       ','.join(str(val) for val in warmup_prompt_length))
    monkeypatch.setenv('VLLM_SPYRE_WARMUP_BATCH_SIZES',
                       ','.join(str(val) for val in warmup_batch_size))

    if all(len(s) == 3 for s in warmup_shapes):
        warmup_new_tokens = [t[1] for t in warmup_shapes]
        monkeypatch.setenv('VLLM_SPYRE_WARMUP_NEW_TOKENS',
                           ','.join(str(val) for val in warmup_new_tokens))


class ModelInfo(NamedTuple):
    name: str
    revision: str | None = None
    is_quantized: bool = False

    def __str__(self):
        return f"ModelInfo({self.name}@{self.revision})"


class RemoteOpenAIServer:
    """Subprocess wrapper that boots a vllm server with `vllm serve` for testing
    against"""

    DUMMY_API_KEY = "token-abc123"  # vLLM's OpenAI server does not need API key

    def __init__(
        self,
        model: str | ModelInfo,
        vllm_serve_args: list[str],
        *,
        env_dict: Optional[dict[str, str]] = None,
        seed: Optional[int] = 0,
        auto_port: bool = True,
        max_wait_seconds: Optional[float] = None,
    ) -> None:
        # NB: This implementation does not ensure that the model is downloaded
        # before booting the server, it should be used with models already
        # cached on disk
        if isinstance(model, ModelInfo):
            if model.revision is not None:
                vllm_serve_args = vllm_serve_args + [
                    "--revision", model.revision
                ]
            model_name = model.name
        else:
            model_name = model

        if auto_port:
            if "-p" in vllm_serve_args or "--port" in vllm_serve_args:
                raise ValueError("You have manually specified the port "
                                 "when `auto_port=True`.")

            # Don't mutate the input args
            vllm_serve_args = vllm_serve_args + [
                "--port", str(get_open_port())
            ]
        if seed is not None:
            if "--seed" in vllm_serve_args:
                raise ValueError("You have manually specified the seed "
                                 f"when `seed={seed}`.")

            vllm_serve_args = vllm_serve_args + ["--seed", str(seed)]

        parser = FlexibleArgumentParser(
            description="vLLM's remote OpenAI server.")
        parser = make_arg_parser(parser)
        args = parser.parse_args(["--model", model_name, *vllm_serve_args])
        self.host = str(args.host or "localhost")
        self.port = int(args.port)

        env = os.environ.copy()
        if env_dict is not None:
            env.update(env_dict)
        self.proc = subprocess.Popen(
            ["vllm", "serve", model_name, *vllm_serve_args],
            env=env,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        max_wait_seconds = max_wait_seconds or 600
        self._wait_for_server(url=self.url_for("health"),
                              timeout=max_wait_seconds)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.shutdown()

    def shutdown(self):
        self.proc.terminate()
        try:
            self.proc.wait(8)
        except subprocess.TimeoutExpired:
            # force kill if needed
            self.proc.kill()

    def _wait_for_server(self, *, url: str, timeout: float):
        # run health check
        start = time.time()
        while True:
            try:
                if requests.get(url).status_code == 200:
                    break
            except Exception:
                # this exception can only be raised by requests.get,
                # which means the server is not ready yet.
                # the stack trace is not useful, so we suppress it
                # by using `raise from None`.
                result = self.proc.poll()
                if result is not None and result != 0:
                    raise RuntimeError("Server exited unexpectedly.") from None

                time.sleep(0.5)
                if time.time() - start > timeout:
                    raise RuntimeError(
                        "Server failed to start in time.") from None

    @property
    def url_root(self) -> str:
        return f"http://{self.host}:{self.port}"

    def url_for(self, *parts: str) -> str:
        return self.url_root + "/" + "/".join(parts)

    def get_client(self, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = 600
        return openai.OpenAI(
            base_url=self.url_for("v1"),
            api_key=self.DUMMY_API_KEY,
            max_retries=0,
            **kwargs,
        )

    def get_async_client(self, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = 600
        return openai.AsyncOpenAI(
            base_url=self.url_for("v1"),
            api_key=self.DUMMY_API_KEY,
            max_retries=0,
            **kwargs,
        )


# get model directory path from env
# if unset, test model paths are assumed to be either hf hub names or absolute
# paths
def get_spyre_model_dir_path() -> Path:
    model_dir_path = os.environ.get("VLLM_SPYRE_TEST_MODEL_DIR", "")
    return Path(model_dir_path)


# add pytest markers to supported different backends
def get_spyre_backend_list():
    backend_list = ["eager", "inductor", "sendnn"]

    backends = []
    for backend in backend_list:
        marks = []
        if backend == "eager":
            marks = [pytest.mark.cpu]
        elif backend == "sendnn":
            marks = [pytest.mark.spyre]

        backends.append(pytest.param(backend, marks=marks, id=backend))
    return backends


# get model names from env, if not set then use default models for each type.
# Multiple models can be specified with a comma separated list in
# VLLM_SPYRE_TEST_MODEL_LIST
def get_spyre_model_list(isEmbeddings=False,
                         isScoring=False,
                         full_size_models=False):
    """Returns a list of pytest.params. The values are NamedTuples with a name
    and revision field."""
    user_test_model_list = os.environ.get("VLLM_SPYRE_TEST_MODEL_LIST")
    if not user_test_model_list:
        return _default_test_models(isEmbeddings, isScoring, full_size_models)

    # User overridden model list
    spyre_model_dir_path = get_spyre_model_dir_path()
    if isEmbeddings:
        marks = [pytest.mark.embedding]
    elif isScoring:
        marks = [pytest.mark.scoring]
    else:
        marks = [pytest.mark.decoder]

    test_model_list = []
    for model in user_test_model_list.split(","):
        model_path = str(spyre_model_dir_path / model.strip())
        test_model_list.append(
            pytest.param(ModelInfo(name=model_path),
                         marks=marks,
                         id=model.strip()))
    return test_model_list


def _default_test_models(isEmbeddings=False,
                         isScoring=False,
                         full_size_models=False):
    """Return the default set of test models as pytest parameterizations"""
    if isEmbeddings:
        model = ModelInfo(name="sentence-transformers/all-roberta-large-v1",
                          revision="cf74d8acd4f198de950bf004b262e6accfed5d2c")
        return [
            pytest.param(model, marks=[pytest.mark.embedding], id=model.name)
        ]

    if isScoring:
        model = ModelInfo(name="cross-encoder/stsb-roberta-large",
                          revision="2b12c2c0088918e76151fd5937b7bba986ef1f98")
        return [
            pytest.param(model, marks=[pytest.mark.scoring], id=model.name)
        ]

    # Decoders
    # We run tests for both the full-precision bf16 and fp8-quantized models,
    # but by default the `pytest.mark.quantized` marker is de-selected unless
    # the test command includes `-m quantized`.
    if not full_size_models:
        tinygranite = ModelInfo(
            name="ibm-ai-platform/micro-g3.3-8b-instruct-1b",
            revision="6e9c6465a9d7e5e9fa35004a29f0c90befa7d23f")
        tinygranite_fp8 = ModelInfo(
            name="ibm-ai-platform/micro-g3.3-8b-instruct-1b-FP8",
            revision="0dff8bacb968836dbbc7c2895c6d9ead0a05dc9e",
            is_quantized=True)
        params = [
            pytest.param(tinygranite,
                         marks=[pytest.mark.decoder],
                         id=tinygranite.name),
            pytest.param(tinygranite_fp8,
                         marks=[pytest.mark.decoder, pytest.mark.quantized],
                         id=tinygranite_fp8.name)
        ]
        return params

    # Full-size decoders
    granite = ModelInfo(name="ibm-granite/granite-3.3-8b-instruct",
                        revision="51dd4bc2ade4059a6bd87649d68aa11e4fb2529b")
    granite_fp8 = ModelInfo(
        name="ibm-granite/granite-3.3-8b-instruct-FP8",
        revision="4b5990b8d402a75febe0086abbf1e490af494e3d")
    params = [
        pytest.param(granite, marks=[pytest.mark.decoder], id=granite.name),
        pytest.param(granite_fp8,
                     marks=[pytest.mark.decoder, pytest.mark.quantized],
                     id=granite_fp8.name)
    ]
    return params


def create_text_prompt(model: ModelInfo, min_token_length: int,
                       max_token_length: int) -> str:
    """Create a text prompt for the specified model that will tokenize to within
    the specified token length range."""
    tokenizer = AutoTokenizer.from_pretrained(model.name,
                                              revision=model.revision)
    pepper = "🌶️"
    pepper_tokens = len(tokenizer.encode(pepper, add_special_tokens=False))

    # Find a good starting number of peppers
    prompt = pepper * (min_token_length // pepper_tokens + 1)

    # And add more until we're over the minimum token length
    while len(tokenizer.encode(prompt)) <= min_token_length:
        prompt += pepper

    # Make sure this prompt is within the specified range
    assert min_token_length < len(tokenizer.encode(prompt)) < max_token_length

    return prompt


def create_seq_prompt(model: ModelInfo, token_length: int) -> str:
    """Create a repeating sequential number prompt for the specified
    model that will tokenize to exactly the specified token length."""

    tokenizer = AutoTokenizer.from_pretrained(model.name,
                                              revision=model.revision)

    # 20-token pattern
    pattern = "0 1 2 3 4 5 6 7 8 9 "

    # Repeat to token_length
    repeat_count = (token_length // 20) + 1
    text_prompt = pattern * repeat_count

    # Tokenize and slice
    tokens = tokenizer.encode(text_prompt)[:token_length]

    # Assert exact token length
    assert len(tokens) == token_length, \
        f"Token length mismatch: {len(tokens)} != {token_length}"

    return tokenizer.decode(tokens)


def create_random_request(
    request_id: int,
    num_tokens: int,
    sampling_params: SamplingParams,
    from_model_vocab: bool = False,
    model: Optional[ModelInfo] = None,
) -> Request:

    tokenizer = AutoTokenizer.from_pretrained(model.name,
                                              revision=model.revision)
    if from_model_vocab:
        assert model is not None, "Prompt requested to be generated from " \
        "model's vocabulary: need to provide model."

        valid_token_ids = sorted([
            v for v in tokenizer.vocab.values()
            if v not in tokenizer.all_special_ids
        ])
        prompt_token_ids = random.choices(valid_token_ids, k=num_tokens)
    else:
        # start with existing prompts and tokenize them
        prompts = get_longer_chicken_soup_prompts(1)
        tokenized_prompts = tokenizer(prompts)["input_ids"]
        prompt_token_ids = [p[:num_tokens] for p in tokenized_prompts][0]

        # make sure we get enough tokens from the prompts
        assert (len(prompt_token_ids) == num_tokens
                ), f"need {num_tokens} but got {len(prompt_token_ids)}"

    return Request(request_id=str(request_id),
                   prompt_token_ids=prompt_token_ids,
                   sampling_params=sampling_params,
                   eos_token_id=None,
                   arrival_time=0,
                   lora_request=None,
                   pooling_params=None,
                   cache_salt=None)


def skip_unsupported_tp_size(size: int, backend: str):
    if backend in ["eager", "inductor"]:
        # Spyre cards aren't required for running TP on CPU backends
        # But it's really slow to run tp > 2
        if size > 2:
            pytest.skip("Skipping TP test on CPU with TP size > 2")
        return
    cards = int(os.getenv("AIU_WORLD_SIZE", "0"))
    if cards < size:
        pytest.skip(f"Cannot run TP size {size}: "
                    f"only {cards} cards are available")


def get_chicken_soup_prompts(num_prompts: int) -> list[str]:
    template = (
        "Below is an instruction that describes a task. Write a response that "
        "appropriately completes the request. Be polite in your response to the"
        " user.\n\n### Instruction:\n{}\n\n### Response:")

    prompts = [
        template.format("Provide a list of instructions "
                        "for preparing chicken soup."),
        template.format("Provide me a list of things that I can do with my "
                        "new found wealth."),
        template.format(
            "how do I add multiple new columns in m for power query or \
                power bi?"),
        template.format("Convert char to string in Java."),
    ]

    if num_prompts > 4:
        prompts = prompts * (math.ceil(num_prompts / 4))

    return prompts[:num_prompts]


def get_longer_chicken_soup_prompts(num_prompts: int) -> list[str]:
    template = (
        "Below is an instruction that describes a task. Write a response that "
        "appropriately completes the request. Be polite in your response to the"
        " user.\n\n### Instruction:\n{}\n\n### Response:")

    prompts = [
        template.format("Provide a list of instructions "
                        "for preparing chicken soup along with "
                        "rice curry to go with it so that "
                        "the flavor is amazing and make sure to follow the "
                        "recipe that my mum used to make during my "
                        "childhood so that I can relive my good "
                        "memories thanks"),
        template.format("Provide me a list of things that I can do with my "
                        "new found wealth which I have obtained through "
                        "nefarious activities including gambling "
                        "and betting on sports thanks"),
        template.format(
            "how do I add multiple new columns in m for power query or \
                power bi? Can you explain that to me like I'm 5 years old "
            "with thorough step by step explanation and covering all edge "
            "cases thanks"),
        template.format(
            "Convert char to string in Java "
            "and write unit tests for the same, making sure they all pass "
            "and we get amazing test coverage along with high level "
            "correctness so that the PR reviewers have an easy time "
            "reviewing the changes thanks"),
    ]

    if num_prompts > 4:
        prompts = prompts * (math.ceil(num_prompts / 4))

    return prompts[:num_prompts]


def write_sample_model_config(tmp_path,
                              data,
                              filename="model_compile.log.json"):
    """Helper to write a sample model_compile.log.json in tmp_path."""
    config_path = tmp_path / filename
    config_path.write_text(json.dumps(data))
    return config_path



import contextlib
import math

import torch
from vllm.logger import init_logger

logger = init_logger(__name__)


@contextlib.contextmanager
def stagger_region(limit: int, world_size: int, rank: int):
    """
    Limit the number of concurrent processes into this region of code.
    Processes yield from this function when they are allowed to enter the
    region of code. Processes return from this function when all of the
    processes have completed the region of code.

    :param limit: Number of concurrent processes allowed if > 0.
    :param world_size: Total world size, usually TP * PP.
    :param rank: Rank of calling worker process.
    """
    if limit > 0 and limit < world_size:
        for _set in range(math.ceil(world_size / float(limit))):
            if rank < (_set + 1) * limit:
                break
            torch.distributed.barrier()
        logger.info("Stagger Region Enter (Set: %d) of %d", _set + 1,
                    math.ceil(world_size / float(limit)))
    yield {}

    # TODO: make sure this isn't called excessively

    if limit > 0 and limit < world_size:
        logger.info("Rank %d Done With Stagger Region", rank)
        for _set in range(math.ceil(world_size / float(limit))):
            if rank >= (_set + 1) * limit:
                continue
            torch.distributed.barrier()
        logger.info("Stagger Region: All Complete")



# SPDX-License-Identifier: Apache-2.0

import math
import os
from collections import deque
from typing import TYPE_CHECKING

from vllm.logger import init_logger
from vllm.v1.core.sched.scheduler import Scheduler
from vllm.v1.engine import EngineCoreOutputs
from vllm.v1.outputs import ModelRunnerOutput
from vllm.v1.request import Request

import vllm_spyre.envs as envs_spyre
from vllm_spyre.platform import SpyrePlatform
from vllm_spyre.v1.worker.spyre_model_runner import CBSpyreModelRunnerOutput

if TYPE_CHECKING:
    from vllm.v1.core.sched.output import SchedulerOutput
else:
    SchedulerOutput = None

logger = init_logger(__name__)

NO_WARMUP_FIT_STOP_REASON = "Request did not fit any warmup shape"


class SpyreScheduler(Scheduler):
    """Base class inheriting from the V1 scheduler to support static 
    and continuous batching respecting AIU Spyre constraints."""

    def __init__(self, *args, **kwargs) -> None:
        # Initialize vLLM scheduler
        super().__init__(*args, **kwargs)


class StaticBatchingSpyreScheduler(SpyreScheduler):
    """ Support of static batching """

    def __init__(self, *args, **kwargs) -> None:
        # Initialize SpyreScheduler
        super().__init__(*args, **kwargs)

        # Add our own state for handling Spyre constraints:
        # all warmup shapes that we can support
        self.spyre_warmup_shapes: tuple[dict[str, int], ...] = \
            SpyrePlatform.get_warmup_shapes(self.scheduler_config)

    def schedule(self) -> SchedulerOutput:
        """This override adds constraints and then delegates most of the work
        to the base scheduler"""
        # First purge the full waiting queue into our holdback queue, preserving
        # priority, so that the base scheduler does not see them.
        # This lets us ensure that the set of requests scheduled have at least
        # one common warmup shape.
        holdback_queue: deque[Request] = deque()
        while self.waiting:
            holdback_queue.append(self.waiting.popleft())

        # store requests which don't fit the warmup shapes of the current batch
        skip_queue: deque[Request] = deque()

        # If no requests are currently running, we can now release requests back
        # into the waiting queue in priority order for the scheduler to prefill.
        # These must share a common warmup shape
        if len(self.running) == 0:

            # Make a copy of the warmup shapes
            available_warmup_shapes = list(self.spyre_warmup_shapes)

            while holdback_queue:
                request = holdback_queue[0]

                # prune the possible shapes to only those that fit this request
                # and the growing batch size
                available_warmup_shapes = self._get_matching_warmup_shapes(
                    request=request,
                    warmup_shapes=available_warmup_shapes,
                    current_batch_size=len(self.waiting))

                if len(available_warmup_shapes) > 0:
                    # There is still at least one valid shape, so add to the
                    # waiting queue
                    self.waiting.append(holdback_queue.popleft())
                    # remember the available warmup shapes of the current batch
                    last_available_warmup_shapes = available_warmup_shapes
                else:
                    # calculating the max possible batch size among the
                    # available warmup shapes of the scheduled requests
                    max_batch = max([
                        d['batch_size'] for d in last_available_warmup_shapes
                    ])

                    # if there is potential space in the batch but the current
                    # request does not fit, skip it and try with the next
                    if len(self.waiting) < max_batch:
                        available_warmup_shapes = last_available_warmup_shapes
                        skip_queue.append(holdback_queue.popleft())
                    else:
                        # If the batch is full, we exit the loop here
                        break

            logger.debug(
                "Scheduling a new batch of %d requests, holding back %d "
                "requests", len(self.waiting), len(holdback_queue))
        else:
            logger.debug("Scheduling a running batch of %d requests",
                         len(self.running))

        # delegate to super of SpyreScheduler: base V1 Scheduler
        outputs = super(SpyreScheduler, self).schedule()

        # first move skipped and then unscheduled requests back
        # to the waiting queue, preserving priority
        while skip_queue:
            self.waiting.append(skip_queue.popleft())

        while holdback_queue:
            self.waiting.append(holdback_queue.popleft())

        return outputs

    def _get_matching_warmup_shapes(
            self, request: Request, warmup_shapes: list[dict[str, int]],
            current_batch_size: int) -> list[dict[str, int]]:
        """Return the subset of shapes that match this request"""
        max_tokens = 0
        if request.sampling_params is not None and\
                request.sampling_params.max_tokens is not None:
            max_tokens = request.sampling_params.max_tokens

        return [
            shape for shape in warmup_shapes
            if request.num_prompt_tokens <= shape['prompt_length']
            and max_tokens <= shape['new_tokens']
            and current_batch_size < shape['batch_size']
        ]


class ContinuousBatchingSpyreScheduler(SpyreScheduler):
    """ Support of continuous batching """

    # inherited from V1 base scheduler but mypy needs to know the type
    running: list[Request]

    def __init__(self, *args, **kwargs) -> None:
        # Initialize SpyreScheduler
        super().__init__(*args, **kwargs)
        self.tkv = 0
        self.n_free_blocks = 0
        self.block_size = SpyrePlatform.get_block_size()
        self.max_batch_tkv_limit = os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT",
                                             default='-1')
        assert self.max_batch_tkv_limit != '-1', (
            "Expecting the env var VLLM_DT_MAX_BATCH_TKV_LIMIT to be set in "
            "platform.py")
        # cache for self.check_batch_tkv_limit() outer key: tuple(request_ids),
        # inner key: (request_id, max_batch_tkv_limit), value: (lower, upper)
        self._cache_check_batch_tkv_limit: dict[tuple, dict[tuple, tuple]] = {}

    def update_from_output(
        self,
        scheduler_output: SchedulerOutput,
        model_runner_output: ModelRunnerOutput,
    ) -> dict[int, EngineCoreOutputs]:
        # Need an instance of CBSpyreModelRunnerOutput which holds the tkv value
        assert isinstance(model_runner_output, CBSpyreModelRunnerOutput), (
            "Expecting an instance of CBSpyreModelRunnerOutput when doing "
            "continuous batching.")
        self.tkv = model_runner_output.tkv
        self.n_free_blocks = model_runner_output.n_free_blocks
        return super(SpyreScheduler,
                     self).update_from_output(scheduler_output,
                                              model_runner_output)

    def schedule(self) -> "SchedulerOutput":
        """This override adds constraints and then delegates most of the work
        to the base scheduler

        To avoid additional specialization, some requests are held back from the
        base scheduler but are restored after.
        """
        # First purge the full waiting queue into our holdback queue, preserving
        # priority, so that the base scheduler does not see them.
        holdback_queue: deque[Request] = deque()
        while self.waiting:
            holdback_queue.append(self.waiting.popleft())

        # Check if new requests can be scheduled.
        while holdback_queue:
            if self.can_schedule(holdback_queue[0]):
                # Add request to the waiting queue
                self.waiting.append(holdback_queue.popleft())
            else:
                # Otherwise, we simply stop here so that the scheduler
                # can work with the batch we have
                break

        # Schedule Prefill and Decode separately
        if len(self.waiting) > 0:
            # For prefill, hide current decodes from the scheduler
            running_holdback = self.running
            self.running = []
            logger.debug(
                "Scheduling a prefill step of %d requests, holding back %d "
                "requests", len(self.waiting), len(holdback_queue))
        else:
            running_holdback = []
            logger.debug("Scheduling a decode step of %d requests",
                         len(self.running))

        # delegate to super of SpyreScheduler: base V1 Scheduler
        outputs = super(SpyreScheduler, self).schedule()

        # restore holdbacks after running the base scheduler
        self.running = self.running + running_holdback
        while holdback_queue:
            self.waiting.append(holdback_queue.popleft())

        return outputs

    def can_schedule(self, request) -> bool:
        max_prompt_batch_size = 1
        max_context_len = self.scheduler_config.max_model_len

        # running and waiting queues are both empty -> start a new batch
        # which can always be scheduled
        if len(self.running) + len(self.waiting) == 0:
            return True

        # check that there is space in the current decode batch
        cond1 = len(self.running) + len(
            self.waiting) < self.max_num_running_reqs
        # check that there is space in the prefill batch
        cond2 = len(self.waiting) < max_prompt_batch_size
        # check that the prompt length does not exceed the current tkv
        cond3 = request.num_prompt_tokens <= self.tkv
        # check that the number of requested tokens can be served
        cond4 = request.max_tokens <= (max_context_len - self.tkv)
        # check that there are enough free blocks/pages remaining
        # Note: we only have to do check in case of a running batches
        # (not start_new_batch), because the minimal number of blocks covers
        # the context length for a single sequence, so tkv < block size is ok
        num_blocks_required = math.ceil(
            (self.tkv + request.max_tokens - 1) / self.block_size)
        # optimization: subtract the padding blocks from the reserved blocks
        num_fully_padded_blocks = math.floor(
            (self.tkv - request.num_prompt_tokens) / self.block_size)
        num_blocks_required -= num_fully_padded_blocks
        cond5 = num_blocks_required <= self.n_free_blocks
        # scheduling heuristic: prefill vs decode prioritization
        # note that prefills are performed on the minimal number of blocks
        # needed and prefill time is thus proportional to the number of blocks
        num_blocks_prefill = math.ceil(
            self.tkv / self.block_size) - num_fully_padded_blocks
        # if VLLM_SPYRE_N_TOKENS_PREFILL_PRIO is -1 -> no heuristic is enforced
        cond6 = (num_blocks_prefill * self.block_size
                 <= envs_spyre.VLLM_SPYRE_N_TOKENS_PREFILL_PRIO) if (
                     envs_spyre.VLLM_SPYRE_N_TOKENS_PREFILL_PRIO
                     >= 0) else True
        # check that batch size x tkv is smaller than the max supported number
        cond7 = lambda: self.check_batch_tkv_limit(request=request,
                                                   tkv=self.tkv,
                                                   running=self.running,
                                                   max_batch_tkv_limit=self.
                                                   max_batch_tkv_limit)

        if cond1 and cond2 and cond3 and cond4 and cond5 and cond6 and cond7():
            return True

        # the following conditions must always be true, if not we can exit here
        if not (cond1 and cond2 and cond4 and cond5 and cond6 and cond7()
                ) or not envs_spyre.VLLM_SPYRE_ENABLE_PREFILL_OPTIMIZATION:
            return False

        # cond3 is violated: request.num_prompt_tokens > self.tkv
        # check whether the new sequence can join the decode batch by
        # increasing the current tkv by a multiple of the block size
        tkv_offset = math.ceil((request.num_prompt_tokens - self.tkv) /
                               self.block_size) * self.block_size
        tkv_updated = self.tkv + tkv_offset
        # check cond4 again with updated tkv for current sequence
        cond4_updated = request.max_tokens <= (max_context_len - tkv_updated)

        # check cond4 for all other sequences in the current decode batch
        for req in self.running:
            cond4_current = req.max_tokens <= (max_context_len - tkv_updated)
            cond4_updated = cond4_updated and cond4_current
            # early exiting loop if violated 4th condition
            if not cond4_updated:
                return False

        # check if enough number of blocks to serve sequence with updated tkv
        num_blocks_required_updated = math.ceil(
            (tkv_updated + request.max_tokens - 1) / self.block_size)
        cond5_updated = num_blocks_required_updated <= self.n_free_blocks

        # check prefill vs decode prioritization with updated tkv
        # Note: num_fully_padded_blocks is always 0 in this code branch by
        # construction: if the prompt is bigger than self.tkv, we shift
        # self.tkv by tkv_offset to just accommodate the new prompt. The
        # alignment with self.tkv this will require max block_size - 1 pads.
        num_blocks_prefill_updated = math.ceil(tkv_updated / self.block_size)
        cond6_updated = (num_blocks_prefill_updated * self.block_size
                         <= envs_spyre.VLLM_SPYRE_N_TOKENS_PREFILL_PRIO) if (
                             envs_spyre.VLLM_SPYRE_N_TOKENS_PREFILL_PRIO
                             >= 0) else True

        # check that batch size x tkv is smaller than the max supported number
        # with updated tkv (cond6) -> only call if the other cond are met
        cond7_updated = lambda: self.check_batch_tkv_limit(
            request=request,
            tkv=tkv_updated,
            running=self.running,
            max_batch_tkv_limit=self.max_batch_tkv_limit)

        return (cond4_updated and cond5_updated and cond6_updated
                and cond7_updated())

    def check_batch_tkv_limit(self, request, tkv, running,
                              max_batch_tkv_limit) -> bool:
        """
        Check whether adding a new sequence to the decode batch would violate
        Spyre's maximum batch volume constraint.

        In Spyre, the product of `batch_size` and the current `tkv` 
        (tokens-per-sequence) must not exceed the limit defined by 
        `VLLM_DT_MAX_BATCH_TKV_LIMIT`. Before scheduling a new sequence, 
        we must ensure that this constraint will hold for all decoding 
        steps that result from combining the new sequence with the currently 
        running decode batch.

        This implementation:
        1. Computes the maximum possible `tkv` for each sequence in the 
        decode batch.
        2. Sorts these values in ascending order.
        3. Iterates through them, stopping once the `tkv` of the new sequence.
        is reached. Remaining sequences do not need to be checked explicitly, 
        since they were validated when they were added (by inductive reasoning).

        Note: drawing explaining the algorithm in more detail uploaded here: 
        https://github.com/vllm-project/vllm-spyre/pull/363#issuecomment-3173605517
        """
        # checking if cached result can be used
        outer_key = tuple(r.request_id
                          for r in running)  # decode batch changes
        inner_key = (request.request_id, max_batch_tkv_limit
                     )  # new request changes
        cache = self._cache_check_batch_tkv_limit
        if (outer_key in cache) and (inner_key in cache[outer_key]):
            (lower, upper) = cache[outer_key][inner_key]
            if tkv <= lower or tkv >= upper:
                logger.debug(
                    "Cache hit function check_batch_tkv_limit: returning %s",
                    str(tkv <= lower))
                return tkv <= lower

        # Compute the effective token length of the new request
        new_req_tkv = tkv + request.max_tokens - 1

        # Compute token lengths for all running requests (decode batch)
        decode_req_tkvs = [
            tkv + req.max_tokens - 1 -
            (req.num_computed_tokens - req.num_prompt_tokens)
            for req in running
        ]
        # Sort decode requests token lengths in ascending order
        decode_req_tkvs.sort()

        # Initialize values
        batch_size = len(running) + 1
        max_batch_tkv = 0

        # Try adding the new request to the batch and check the max volume
        for decode_req_tkv in decode_req_tkvs:
            if new_req_tkv <= decode_req_tkv:
                # If the new request is shorter, it limits the batch volume
                max_batch_tkv = max(max_batch_tkv, batch_size * new_req_tkv)
                break
            else:
                # Otherwise, use the current (longer) request's volume
                max_batch_tkv = max(max_batch_tkv, batch_size * decode_req_tkv)
                # decrease batch_size by 1 as the current request finished
                batch_size -= 1

        return_value = max_batch_tkv <= int(max_batch_tkv_limit)

        if outer_key in cache:
            # decode batch has not changed
            if inner_key not in cache[outer_key]:
                # adding new request to present decode batch
                cache[outer_key][inner_key] = (-math.inf, math.inf)
        else:
            # decode batch has changed, empty the cache in place
            cache.clear()
            cache[outer_key] = {inner_key: (-math.inf, math.inf)}
            logger.debug(
                "Cleared cache of function check_batch_tkv_limit as the " \
                "decode batch has changed."
            )

        # update lower bound (of acceptance) and upper bound (of rejection)
        (lower, upper) = cache[outer_key][inner_key]
        if return_value:
            lower = max(lower, tkv)
        else:
            upper = min(upper, tkv)
        assert lower < upper
        cache[outer_key][inner_key] = (lower, upper)
        logger.debug("Saved cache of function check_batch_tkv_limit: %s",
                     self._cache_check_batch_tkv_limit[outer_key][inner_key])

        return return_value



#!/usr/bin/env python3
"""
Example usage:
python3 container-scripts/spyre_vllm_benchmark.py 
--prompt-dir $HOME/prompts/ 
--tokenizer-dir $HOME/models/granite-3.3-8b-instruct 
--output-dir $HOME/output/ 
--port 8000 
--max-tokens 64 
--min-tokens 64
"""

# Imports
import argparse
import time
from datetime import datetime
from pathlib import Path
from typing import NamedTuple

import requests
from openai import APIConnectionError, OpenAI
from transformers import AutoTokenizer, PreTrainedTokenizer


# Classes
class InferenceResults(NamedTuple):
    outputs: list[str]
    inference_time: float
    output_token_count: int
    ttft: float
    token_latencies: list[list[float]]


# Functions
def parse_args():
    parser = argparse.ArgumentParser(
        description="VLLM Spyre inference benchmarking script.")
    parser.add_argument("--prompt-dir",
                        required=True,
                        type=str,
                        help="Path to directory containing .txt files")
    parser.add_argument("--tokenizer-dir",
                        required=True,
                        type=str,
                        help="Path to a directory containing a tokenizer")
    parser.add_argument("--port",
                        required=False,
                        help="Port of running container to connect to.",
                        default=8000)
    parser.add_argument("--max-tokens",
                        required=False,
                        type=int,
                        help="Maximum number of tokens to generate",
                        default=64)
    parser.add_argument("--min-tokens",
                        required=False,
                        type=int,
                        help="Minimum number of tokens to generate",
                        default=0)
    parser.add_argument(
        "--output-dir",
        required=False,
        type=Path,
        help="Output directory to dump results and performance metrics",
        default=None)
    return parser.parse_args()


def create_client(api_key: str, base_url: str) -> OpenAI:
    """
    Creates and returns an OpenAI client.

    Args:
        api_key (str): The OpenAI API key. 
                       Often set to "EMPTY" for local inference setups.
        base_url (str): The base URL of the OpenAI-compatible API,
                        e.g., "http://localhost:8000/v1".

    Returns:
        OpenAI: An instance of the OpenAI client initialized with the provided
                API key and base URL.
    """
    client = OpenAI(
        api_key=api_key,
        base_url=base_url,
    )
    return client


def test_server_connection(client: OpenAI, endpoint: str) -> bool:
    """
    Tests the connection to a specified endpoint of the OpenAI server.

    Args:
        client (OpenAI): The OpenAI client instance.
        endpoint (str): The relative endpoint to test (e.g., "/models/").

    Returns:
        bool: True if the server responds with a 200 status code; 
              False otherwise.
    """
    try:
        base_url = str(client.base_url).rstrip("/")
        response = requests.get(base_url + endpoint)
        return response.status_code == 200
    except requests.RequestException as e:
        print(e)
        return False


def connect(client: OpenAI, endpoint: str, max_tries: int = 5) -> None:
    """
    Attempts to connect to the specified server endpoint max_tries times.

    Args:
        client (OpenAI): The OpenAI client instance.
        endpoint (str): The relative endpoint to connect to (e.g., "/models").
        max_tries (int): Maximum number of connection attempts. Default is 5.

    Returns:
        None

    Raises:
        RuntimeError: If connection fails after max_tries attempts.
    """
    tries = 0
    while tries < max_tries:
        try:
            base_url = str(client.base_url).rstrip("/")
            address = base_url + endpoint
            response = requests.get(address)
            if response.status_code == 200:
                return
        except requests.RequestException as e:
            print(f"Connection attempt {tries + 1} failed: {e}")
        time.sleep(1)
        tries += 1
    raise RuntimeError(f"Failed to connect to {endpoint} after"
                       f" {max_tries} attempts.")


def get_tokenizer(model_path: str):
    """
    Loads and returns a tokenizer from the specified model path.

    Args:
        model_path (str): Path to the pretrained model directory 
                          or identifier from Hugging Face Hub.

    Returns:
        PreTrainedTokenizer: A tokenizer instance loaded from model_path.
    """
    tokenizer = AutoTokenizer.from_pretrained(model_path)
    return tokenizer


def get_model_from_server(client: OpenAI) -> str:
    """
    Retrieves the first available model ID from the OpenAI-compatible server.

    Args:
        client (OpenAI): An instance of the OpenAI client.

    Returns:
        str: The ID of the first model returned by the server.

    Raises:
        SystemExit: If there is a connection error while fetching models.
    """
    model = None
    try:
        models = client.models.list()
        model = models.data[0].id
        print(f"Found Model: {model}")
    except APIConnectionError as e:
        print(f"Connection Error: {e}")
        exit(1)
    return model


def process_input_prompts(prompt_dir: str) -> list[Path]:
    """
    Collects all `.txt` prompt files from the specified directory.

    Args:
        prompt_dir (str): Path to the directory containing prompt files.

    Returns:
        list[Path]: List of Paths to the `.txt` prompt files found in 
                    the directory.
    """
    prompt_list = list(Path(prompt_dir).glob("*.txt"))
    if not prompt_list:
        print(f"No .txt files found in {prompt_dir}.")
        exit(1)
    print(f"Found {len(prompt_list)} prompt files at {prompt_dir}")
    return prompt_list


def save_results(output_dir: Path, prompt_files: list[Path], model: str,
                 results: InferenceResults):
    """
    Saves model inference outputs and performance metrics to the specified 
    output directory.

    Each prompt's generated output is written to a separate text file named 
    after the prompt, and performance metrics are written to a single 
    `performance_metrics.txt` file.

    Args:
        output_dir (Path): The directory in which to save the output files. 
                           Created if it doesn't exist.
        prompt_files (list[Path]): A list of prompt file paths that were 
                                   used for inference.
        results (InferenceResults): An object containing model outputs 
                                    and performance metrics.

    Returns:
        None

    Raises:
        SystemExit: If the output directory could not be created.
    """
    # Attempt to create output directory
    try:
        output_dir.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        print(f"Failed to create output directory at {output_dir}: {e}")
        exit(1)

    # Write inference outputs
    for file, result in zip(prompt_files, results.outputs):
        with open(output_dir / f"{file.stem}.txt", "w") as f:
            f.write(result)

    # Generate timestamped filename
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    metrics_filename = output_dir / f"performance_metrics_{timestamp}.txt"

    # Write performance metrics
    with open(metrics_filename, "w") as f:
        f.write(f"Results for inference with model: {model}\n")
        f.write(f"Inference Time: {results.inference_time:.4f}s\n")
        f.write(f"TTFT: {results.ttft:.4f}s\n")
        f.write(f"Inference Time w/o TTFT: "
                f"{results.inference_time - results.ttft:.4f}s\n")
        f.write(f"Number of Output Tokens Generated: "
                f"{results.output_token_count} tokens\n")
        f.write(f"Throughput: "
                f"{(results.output_token_count / results.inference_time):.4f}"
                f"tok/s\n")
        f.write("\n== Per-Prompt Performance Metrics ==\n")
        for i, latencies in enumerate(results.token_latencies):
            min_itl = min(latencies)
            max_itl = max(latencies)
            avg_itl = sum(latencies) / len(latencies)
            f.write(f"Prompt {i} ITL (min, max, avg): "
                    f"{min_itl:.4f}s, {max_itl:.4f}s, {avg_itl:.4f}s\n")

    print(f"Saved results to {output_dir}")


def run_inference(client: OpenAI, model: str, tokenizer: PreTrainedTokenizer,
                  prompt_files: list[Path], max_tokens: int,
                  min_tokens: int) -> InferenceResults:
    """
    Runs inference using an OpenAI-compatible client on a set of text prompts.

    This function reads prompt files, tokenizes the inputs, 
    sends them to the server for streamed completion, 
    and calculates performance metrics such as inference time, 
    time to first token (TTFT), and inter-token latency (ITL).

    Args:
        client (OpenAI): An instance of the OpenAI client.
        model (str): The model ID to use for inference.
        tokenizer (PreTrainedTokenizer): The tokenizer used to 
                                         compute token counts.
        prompt_files (list[Path]): A list of file paths pointing to `.txt` 
                                   prompt files.
        max_tokens (int): Maximum number of tokens to generate per prompt.
        min_tokens (int): Minimum number of tokens to generate per prompt.

    Returns:
        InferenceMetrics:
            - outputs (list[str]): Raw list of generated text completions 
                                   for each prompt.
            - inference_time (float): Total time taken for 
                                      inference (seconds).
            - inference_time_no_ttft (float): Time taken for inference 
                                              excluding ttft (seconds).
            - output_token_count (int): Total number of output tokens 
                                        generated across all prompts.
            - ttft (float): Time to first token (seconds).
            - itl (float): Inter-token latency (seconds per token).
    
    Raises:
        Exception: If error occurs during the inference process.
    """
    # Read text from each prompt file
    prompts = [p.read_text() for p in prompt_files]
    # Get token count for each prompt
    for i, (prompt_text, prompt_file) in enumerate(zip(prompts, prompt_files)):
        token_count = len(tokenizer(prompt_text)["input_ids"])
        print(f"Prompt file: {prompt_file.name} "
              f"| Prompt #{i} token count: {token_count}")

    # Single prompt test run
    print("Starting single prompt test run")
    test_prompt = prompts[0]
    try:
        test_response = client.completions.create(
            model=model,
            prompt=test_prompt,
            max_tokens=max_tokens,
            stream=True,
            temperature=0.0,
            extra_body=dict(min_tokens=min_tokens))

        output = [""]
        for chunk in test_response:
            idx = chunk.choices[0].index
            output[idx] += chunk.choices[0].text
    except Exception as e:
        print("Error during single prompt test run:\n")
        print(e)
        exit(1)
    print("Completed single prompt test run")

    print("Starting inference")
    try:
        # Submit inference payload
        start_time = time.perf_counter()
        response = client.completions.create(
            model=model,
            prompt=prompts,
            max_tokens=max_tokens,
            stream=True,
            temperature=0.0,
            extra_body=dict(min_tokens=min_tokens))

        # Collect streamed tokens
        outputs = [""] * len(prompts)
        ttft = None
        last_token_time: list[float  # type: ignore
                              | None] = [None] * len(prompts)
        token_latencies: list[list[float]] = [[] for _ in prompts]
        for chunk in response:
            idx = chunk.choices[0].index
            now = time.perf_counter()

            # Record the TTFT
            if ttft is None:
                ttft = now - start_time

            # Record subsequent ITLs per prompt
            if last_token_time[idx] is not None:
                token_latencies[idx].append(now - last_token_time[idx])

            # Update last‐seen and accumulate text
            last_token_time[idx] = now
            outputs[idx] += chunk.choices[0].text
        end_time = time.perf_counter()
        print("Inference complete")

        # Calculate results
        inference_time = end_time - start_time
        output_token_count = sum(
            len(tokenizer(output)["input_ids"]) for output in outputs)

    except Exception as e:
        print("Error during inference:\n")
        print(e)
        exit(1)

    return InferenceResults(
        outputs,
        inference_time,
        output_token_count,
        ttft,  # type: ignore
        token_latencies)


def main():
    # Collect command line arguments
    args = parse_args()
    tokenizer_dir = args.tokenizer_dir
    port = args.port
    prompt_dir = args.prompt_dir
    max_tokens = args.max_tokens
    min_tokens = args.min_tokens
    output_dir = args.output_dir

    client = create_client("EMPTY", f"http://localhost:{port}/v1")

    # If a server connection is made
    if test_server_connection(client, "/models/"):
        # Prepare model and prompts
        prompt_list = process_input_prompts(prompt_dir)
        tokenizer = get_tokenizer(tokenizer_dir)
        model = get_model_from_server(client)

        # Inference step
        results = run_inference(client, model, tokenizer, prompt_list,
                                max_tokens, min_tokens)

        # Print results
        for file, result in zip(prompt_list, results.outputs):
            print(f"\n== Output for {file.name} ==\n{result}\n")
        print("\n== Inference Performance Metrics ==")
        print(f"Inference Time: {results.inference_time:.4f}s")
        print(f"TTFT: {results.ttft:.4f}s")
        print(f"Inference Time w/o TTFT: "
              f"{results.inference_time - results.ttft:.4f}s")
        print(f"Number of Output Tokens Generated: "
              f"{results.output_token_count} tokens")
        print(f"Throughput: "
              f"{(results.output_token_count / results.inference_time):.4f}"
              f"tok/s")
        print("\n== Per-Prompt Performance Metrics ==")
        for i, latencies in enumerate(results.token_latencies):
            min_itl = min(latencies)
            max_itl = max(latencies)
            avg_itl = sum(latencies) / len(latencies)
            print(f"Prompt {i} ITL (min, max, avg): "
                  f"{min_itl:.4f}s, {max_itl:.4f}s, {avg_itl:.4f}s")

        # Optionally save results
        if output_dir:
            save_results(output_dir, prompt_list, model, results)
    else:
        print("Server connection failed")
        exit(1)


if __name__ == "__main__":
    main()



import sys

# When running this plugin on a Mac, we assume it's for local development
# purposes. However, due to a compatibility issue with vLLM, which overrides
# the Triton module with a placeholder, vLLM may fail to load on macOS. To
# mitigate this issue, we can safely remove the Triton module (if imported)
# and rely on PyTorch to handle the absence of Triton, ensuring fine execution
# in eager mode.
if sys.platform.startswith("darwin"):
    if sys.modules.get('triton'):
        del sys.modules['triton']

import math
import operator
import os
from typing import TYPE_CHECKING, Union

import torch
from transformers.models.granite import GraniteConfig
from vllm.inputs import ProcessorInputs, PromptType
from vllm.logger import init_logger
from vllm.pooling_params import PoolingParams
from vllm.sampling_params import SamplingParams

if TYPE_CHECKING:
    from vllm.config import ModelConfig, VllmConfig
else:
    ModelConfig = None
    VllmConfig = None
from vllm.platforms import Platform, PlatformEnum

import vllm_spyre.envs as envs_spyre
from vllm_spyre.compilation_utils import handle_disable_compilation

logger = init_logger(__name__)

THREADING_ENVS = [
    "OMP_NUM_THREADS",
    # "TORCHINDUCTOR_COMPILE_THREADS", # vLLM wants this set to 1
    "DT_PARALLEL_THREADS",  # affects the compilation during warmup
    # set these for good measure
    "OPENBLAS_NUM_THREADS",
    "MKL_NUM_THREADS",
]


# Needed by vllm/model_executor/layers/pooler.py:562
# Copied from vllm/utils/__init__.py
class _StreamPlaceholder:

    def __init__(self):
        self.synchronize = lambda: None


class SpyrePlatform(Platform):
    _enum = PlatformEnum.OOT

    # "spyre" device_name no longer worked due to https://github.com/vllm-project/vllm/pull/16464
    device_name: str = "cpu"
    device_type: str = "cpu"
    # compressed-tensors supported by
    # https://github.com/foundation-model-stack/fms-model-optimizer/blob/main/fms_mo/aiu_addons/__init__.py
    supported_quantization: list[str] = ["gptq", "compressed-tensors"]
    _warmup_shapes: tuple[dict[str, int], ...] | None = None
    _block_size: int = 64  # hardcoded Spyre constraint for now
    _config: VllmConfig = None

    # Backend for dynamic compilation ops
    # See vllm batched_count_greater_than method
    simple_compile_backend: str = envs_spyre.VLLM_SPYRE_SIMPLE_COMPILE_BACKEND

    # Needed by vllm/model_executor/layers/pooler.py:562
    current_stream = lambda _: _StreamPlaceholder()

    @classmethod
    def get_device_name(cls, device_id: int = 0) -> str:
        return "spyre"

    @classmethod
    def is_async_output_supported(cls, enforce_eager: bool | None) -> bool:
        """
        Check if the current platform supports async output.
        """
        return False

    @classmethod
    def check_and_update_config(cls, vllm_config: VllmConfig) -> None:
        # 🌶️🌶️🌶️ Patch in our perf logger before the engine is created
        from vllm_spyre.v1.metrics import patch_async_llm_stat_loggers
        patch_async_llm_stat_loggers()

        # In case vllm passes a default vllm_config to us.
        # This happens when get_current_vllm_config is called
        # without setting the vllm config through
        # set_current_vllm_config
        if vllm_config.model_config is None:
            return

        cls._config = vllm_config
        parallel_config = vllm_config.parallel_config
        scheduler_config = vllm_config.scheduler_config
        model_config = vllm_config.model_config
        cache_config = vllm_config.cache_config

        is_decoder = model_config.runner_type == "generate"

        is_pooling = model_config.runner_type == "pooling"

        if not bool(int(os.getenv("VLLM_USE_V1", "1"))):
            raise ValueError("vllm-spyre is only supported with vLLM v1. "
                             "Please set VLLM_USE_V1=1")
        elif not is_decoder and not is_pooling:
            raise ValueError("Only the 'generate' and 'pooling' runners are "
                             "supported")

        if parallel_config.worker_cls == "auto":
            parallel_config.worker_cls = "vllm_spyre.v1.worker."\
                "spyre_worker.SpyreWorker"

        cls._check_threading_config(parallel_config.world_size)

        # set env vars based on the model
        if is_decoder:
            os.environ["FLEX_OVERWRITE_NMB_FRAME"] = "true"
            os.environ["COMPILATION_MODE"] = "offline_decoder"
        if is_pooling:
            os.environ["FLEX_OVERWRITE_NMB_FRAME"] = "false"
            os.environ["COMPILATION_MODE"] = "offline"

        if envs_spyre.VLLM_SPYRE_USE_CB and is_decoder:
            scheduler_config.scheduler_cls = "vllm_spyre.v1.core."\
                "scheduler.ContinuousBatchingSpyreScheduler"
            if envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS:
                raise ValueError("Prompt logprobs not supported with " \
                "continuous batching")
            if (vllm_config.model_config.quantization
                    and vllm_config.scheduler_config.max_num_seqs == 1):
                raise ValueError(
                    "Batch size 1 not supported for fp8 continuous batching.")
        else:
            # Static batching or embedding model.
            # Override --max-num-seqs to the biggest warmup batch size
            # And override --max-model-len to the biggest warmup sequence
            cls._warmup_shapes = None
            spyre_warmup_shapes = cls.get_warmup_shapes(scheduler_config)
            max_batch_size = 0
            max_seq_len = 0
            for shape in spyre_warmup_shapes:
                max_batch_size = max(max_batch_size, shape["batch_size"])
                max_seq_len = max(max_seq_len,
                                  shape["prompt_length"] + shape["new_tokens"])

            if (envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS
                    and max_batch_size > 1):
                raise ValueError(
                    "Prompt logprobs only supported with batch size 1")

            # verify that warmup shapes are not too large
            model_config.get_and_verify_max_len(max_model_len=max_seq_len)

            # override stuff
            model_config.max_model_len = max_seq_len
            scheduler_config.max_num_seqs = max_batch_size

            scheduler_config.scheduler_cls = (
                    "vllm_spyre.v1.core.scheduler."\
                        "StaticBatchingSpyreScheduler")

        # To disable any paged attention ops in the base scheduler, we:
        # - Set the block size (in tokens) to the maximum sequence length
        #       so that the scheduler thinks an entire sequence will fit in
        #       one single block.
        # - Set `max_num_batched_tokens` to the size of a full batch of full
        #       length requests, so that the scheduler will always have token
        #       budget available to schedule a full batch
        if cache_config is not None:
            cache_config.block_size = model_config.max_model_len
            scheduler_config.max_num_batched_tokens = (
                model_config.max_model_len * scheduler_config.max_num_seqs)

        logger.info(
            "Overriding configurations based on warmup shapes. "
            "max_model_len=%d, max_num_seqs=%d, block_size=%d, "
            "max_num_batched_tokens=%d", model_config.max_model_len,
            scheduler_config.max_num_seqs, cache_config.block_size,
            scheduler_config.max_num_batched_tokens)

        # set env vars for torch_sendnn to consume
        os.environ["VLLM_DT_MAX_CONTEXT_LEN"] = str(
            vllm_config.model_config.max_model_len)
        if (envs_spyre.VLLM_SPYRE_USE_CB
                and vllm_config.model_config.max_model_len > 32 * 1024):
            logger.warning(
                'Max context length is too big. Currently only 32K (32768) ' \
                'context length is supported on Spyre for continuous ' \
                'batching. Results might be off!'
            )
        # min value 2 needed for VLLM_DT_MAX_BATCH_SIZE (compiler constraint)
        # Note that we can still have decodes of batch size 1 as the env var
        # only concerns the max batch size.
        os.environ["VLLM_DT_MAX_BATCH_SIZE"] = str(
            max(vllm_config.scheduler_config.max_num_seqs, 2))

        # Hardcode some things for granite-3.3-8b-instruct
        if cls.is_granite_3_8b(vllm_config.model_config):
            cls.configure_granite_3_8b(vllm_config)

        if not os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT"):
            # max product of batch size x tkv supported by the Spyre compiler
            default_max_batch_tkv_limit = \
                vllm_config.model_config.max_model_len * \
                vllm_config.scheduler_config.max_num_seqs

            os.environ["VLLM_DT_MAX_BATCH_TKV_LIMIT"] = str(
                default_max_batch_tkv_limit)
            logger.info("No model / tensor parallel size specific value for " \
            "VLLM_DT_MAX_BATCH_TKV_LIMIT found. Using the default value " \
            "(max_model_len * max_batch_size): %d", default_max_batch_tkv_limit)

        # scheduling heuristic: prefill vs decode prioritization
        if envs_spyre.VLLM_SPYRE_N_TOKENS_PREFILL_PRIO == -1:
            logger.info(
                "Env var VLLM_SPYRE_N_TOKENS_PREFILL_PRIO for prefill/decode "
                "balancing unset. Defaulting to -1, which always prioritizes "
                "prefills (no scheduler heuristic/ balancing at all).")
        else:
            logger.info(
                "Env var VLLM_SPYRE_N_TOKENS_PREFILL_PRIO for prefill/decode "
                "balancing is set to %s. This means that prefills using up to "
                " %s tokens will always be prioritized over decodes.",
                envs_spyre.VLLM_SPYRE_N_TOKENS_PREFILL_PRIO,
                envs_spyre.VLLM_SPYRE_N_TOKENS_PREFILL_PRIO)

        # Compare requested runtime configuration with supported configurations
        # Don't use top-level import to avoid circular import error
        from vllm_spyre.config.runtime_config_validator import (
            validate_runtime_configuration)

        warmup_shape_tuples = [
            (ws['prompt_length'], ws['new_tokens'], ws['batch_size'])
            for ws in cls._warmup_shapes
        ] if cls._warmup_shapes and not envs_spyre.VLLM_SPYRE_USE_CB else None

        validate_runtime_configuration(
            model_config=model_config,
            tp_size=parallel_config.tensor_parallel_size,
            max_model_len=model_config.max_model_len,
            max_num_seqs=scheduler_config.max_num_seqs,
            warmup_shapes=warmup_shape_tuples)

        handle_disable_compilation(vllm_config, is_decoder)

    @classmethod
    def use_all_gather(cls) -> bool:
        """
        Whether to use allgather in LogitsProcessor to gather the logits.
        """
        return True

    @classmethod
    def is_pin_memory_available(cls) -> bool:
        logger.warning("Pin memory is not supported on Spyre.")
        return False

    @classmethod
    def inference_mode(cls):
        """
        Spyre does not support `torch.inference_mode`.
        This allows to fall back to `torch.no_grad` when inference mode is set.
        """
        return torch.no_grad()

    @classmethod
    def get_warmup_shapes(cls, scheduler_config) -> tuple[dict[str, int], ...]:
        if cls._warmup_shapes is not None:
            return cls._warmup_shapes
        # load warmup shapes and sort by "speed"
        wup_prompt_lens = envs_spyre.VLLM_SPYRE_WARMUP_PROMPT_LENS or []
        if not all(pl % 64 == 0 for pl in wup_prompt_lens):
            raise RuntimeError(
                "All values in VLLM_SPYRE_WARMUP_PROMPT_LENS must be multiples "
                "of 64.")

        wup_batch_sizes = envs_spyre.VLLM_SPYRE_WARMUP_BATCH_SIZES or []
        if len(wup_prompt_lens) != len(wup_batch_sizes):
            raise RuntimeError(
                "The lists in VLLM_SPYRE_WARMUP_PROMPT_LENS and "
                "VLLM_SPYRE_WARMUP_BATCH_SIZES must have equal length")
        if scheduler_config.runner_type == "pooling":
            wup_new_tokens = [0] * len(wup_prompt_lens)
        else:
            wup_new_tokens = envs_spyre.VLLM_SPYRE_WARMUP_NEW_TOKENS or []
            if len(wup_new_tokens) != len(wup_prompt_lens):
                raise RuntimeError(
                    "The lists in VLLM_SPYRE_WARMUP_PROMPT_LENS and "
                    "VLLM_SPYRE_WARMUP_NEW_TOKENS must have equal length")

        logger.info("VLLM_SPYRE_WARMUP_PROMPT_LENS = %s", wup_prompt_lens)
        logger.info("VLLM_SPYRE_WARMUP_NEW_TOKENS = %s", wup_new_tokens)
        logger.info("VLLM_SPYRE_WARMUP_BATCH_SIZES = %s", wup_batch_sizes)

        cls._warmup_shapes = tuple(
            sorted([{
                'prompt_length': pl,
                'new_tokens': nt,
                'batch_size': bs
            } for pl, nt, bs in zip(wup_prompt_lens, wup_new_tokens,
                                    wup_batch_sizes)],
                   key=operator.itemgetter('batch_size', 'prompt_length')))
        return cls._warmup_shapes

    @classmethod
    def get_block_size(cls) -> int:
        return cls._block_size

    @classmethod
    def supports_v1(cls, model_config: ModelConfig) -> bool:
        """Returns whether the current platform can support v1 for the supplied
        model configuration.
        """
        return True

    @classmethod
    def validate_request(
        cls,
        prompt: PromptType,
        params: Union[SamplingParams, PoolingParams],
        processed_inputs: ProcessorInputs | None = None,
    ) -> None:
        """Raises if this request is unsupported on this platform"""
        if isinstance(params, PoolingParams):
            # Only validating generation requests for now
            return None

        # Note: Currently prompt logprobs are not supported, therefore
        # envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS is hardcoded to False
        if (params.prompt_logprobs is not None
                and not envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS):
            raise ValueError("Prompt logprobs are currently not supported.")

        if isinstance(prompt, dict) and "prompt_token_ids" in prompt:
            prompt_len = len(prompt["prompt_token_ids"])
        elif processed_inputs is not None:
            if "encoder" in processed_inputs:
                raise ValueError("Encoder-decoder models not supported ")
            prompt_len = len(processed_inputs["prompt_token_ids"])
        else:
            # We need a prompt length to do any validation here
            return

        max_tokens = 0
        if params is not None and params.max_tokens is not None:
            max_tokens = params.max_tokens

        if envs_spyre.VLLM_SPYRE_USE_CB:
            # For continuous batching, check if the request is within the max
            # context length. This needs to take the padded prompt length
            # into account.

            # ceil division to pad to next block boundary
            prompt_padding_len = math.ceil(
                prompt_len / cls._block_size) * cls._block_size
            if (prompt_padding_len + max_tokens
                    > cls._config.scheduler_config.max_model_len):
                raise ValueError(
                    "Could not add request: prompt length is "
                    f"{prompt_len} tokens, which gets padded to "
                    f"{prompt_padding_len} tokens, maximum number of output "
                    f"tokens is {max_tokens} tokens, but max model context "
                    f"length is {cls._config.scheduler_config.max_model_len}.")
        else:
            # For non-continuous batching, check if the request matches a warmup
            # shape
            assert cls._warmup_shapes is not None, "Warmup shapes must be set"
            if len(
                    cls._get_matching_warmup_shapes(
                        prompt_len=prompt_len,
                        max_tokens=max_tokens,
                        warmup_shapes=cls._warmup_shapes)) == 0:
                raise ValueError(
                    "No applicable warmup shape exists for "
                    f"combination of prompt length ({prompt_len} tokens) "
                    "and maximum number of output tokens to be "
                    f"generated ({max_tokens} tokens)")

    @classmethod
    def _get_matching_warmup_shapes(
            cls, prompt_len: int, max_tokens: int,
            warmup_shapes: tuple[dict[str, int], ...]) -> list[dict[str, int]]:
        """Return the subset of shapes that match this request"""
        return [
            shape for shape in warmup_shapes
            if prompt_len <= shape['prompt_length']
            and max_tokens <= shape['new_tokens']
        ]

    @classmethod
    def _check_threading_config(cls, worker_count: int):
        """
        Check parallelism configuration to avoid CPU contention

        Libraries that support multi-threading (eg. OpenMP) default to
        parallelism based on the number of CPUs on the host. This can lead to
        CPU contention in containerized deployments especially when process
        forking is involved. This function provides better default behavior.
        """

        # The quay.io/ibm-aiu/spyre-base image includes shell scripts that
        # automatically set OMP_NUM_THREADS to the result of `nproc --all`.
        #
        # vLLM also already has logic around threading to be aware of,
        #  - sets TORCHINDUCTOR_COMPILE_THREADS=1 (https://github.com/vllm-project/vllm/blob/baba0389f7e810a361fff5229ce20c2d5a2b1fac/vllm/env_override.py#L38-L39)
        #  - it will set OMP_NUM_THREADS=1 when using multiple workers (https://github.com/vllm-project/vllm/blob/baba0389f7e810a361fff5229ce20c2d5a2b1fac/vllm/executor/multiproc_worker_utils.py#L304)
        #  - has configurations for OMP thread binding (https://github.com/vllm-project/vllm/blob/baba0389f7e810a361fff5229ce20c2d5a2b1fac/vllm/envs.py#L435-L438)
        #    - the bind attempts to detect NUMA nodes (https://github.com/vllm-project/vllm/blob/baba0389f7e810a361fff5229ce20c2d5a2b1fac/vllm/v1/worker/cpu_worker.py#L111)

        assert worker_count > 0
        # Always print current env for awareness
        env_map = {env: os.getenv(env) for env in THREADING_ENVS}
        logger.info(
            "Initial threading configurations: %s",
            ' '.join([f"{env}={value}" for env, value in env_map.items()]))

        # Try to determine the CPU time/cores that we are allocated
        cpu_count: float | None = None
        detection_message = ""

        if (num_cpu := envs_spyre.VLLM_SPYRE_NUM_CPUS) > 0:
            cpu_count = num_cpu
            detection_message = f"VLLM_SPYRE_NUM_CPUS is set to {cpu_count}"
        else:
            try:
                # try to query cgroup CPU limits
                with open('/sys/fs/cgroup/cpu.max') as f:
                    quota_str, period_str = f.read().strip().split()

                if quota_str != 'max':
                    quota = int(quota_str)
                    period = int(period_str)
                    cpu_count = quota / period
                    detection_message = \
                        f"Detected cgroup CPU limit of {cpu_count}"

            except FileNotFoundError:
                # file may not exist if not running under cgroups v2
                pass
            except Exception as e:
                logger.debug(
                    "Error parsing /sys/fs/cgroup/cpu.max to get CPU info",
                    exc_info=e)

            # try psutil to get physical core count
            if cpu_count is None:
                try:
                    import psutil
                    cpu_count = float(psutil.cpu_count(logical=False))
                    detection_message = \
                        f"Detected {cpu_count} physical CPUs from " \
                         "psutil.cpu_count(logical=False)"
                except ImportError:
                    logger.info("Install psutil to count physical CPU cores")
                    pass
                except Exception as e:
                    logger.debug("Error using psutil", exc_info=e)

            # could try `nproc` here, but it is affected by
            # OMP_NUM_THREADS itself

            # try os.cpu_count() to get node CPU count
            if cpu_count is None and (cpu_count_res :=
                                      os.cpu_count()) is not None:
                cpu_count = float(cpu_count_res)
                detection_message = \
                    f"Detected {cpu_count} CPUs from `os.cpu_count()`"

        # NOTE: math.ceil can output a number for each worker that sums
        # to a total greater than cpu_count.
        cpus_per_worker = math.ceil(
            cpu_count / worker_count) if cpu_count is not None else None

        thread_warning = "Excessive threads may result in CPU contention. " \
             + "Note that each worker processes has its own thread pools." \
                if worker_count > 1 else ""
        failed_detection_message = "Unable to detect available CPUs to " \
            "validate threading configuration."

        if envs_spyre.VLLM_SPYRE_UPDATE_THREAD_CONFIG:
            if cpus_per_worker is None:
                raise RuntimeError(
                    f"{failed_detection_message} Set VLLM_SPYRE_NUM_CPUS or "
                    "use VLLM_SPYRE_UPDATE_THREAD_CONFIG=0 and configure "
                    "manually.")

            for env in THREADING_ENVS:
                os.environ[env] = str(cpus_per_worker)

            logger.info(
                "%s for %d workers. Since VLLM_SPYRE_UPDATE_THREAD_CONFIG is "
                "enabled, setting threading configurations to %d",
                detection_message, worker_count, cpus_per_worker)
            return

        # In the case that VLLM_SPYRE_UPDATE_THREAD_CONFIG is not enabled,
        # check configs and maybe log a warning
        if cpus_per_worker is None:
            logger.info("%s %s", failed_detection_message, thread_warning)
            return

        def _float_or_0(s: str) -> float:
            try:
                return float(s)
            except ValueError:
                return 0.0

        if any((value is None or _float_or_0(value) > 1.2 * cpus_per_worker)
               for value in env_map.values()):
            logger.warning(
                "%s %s for %d workers. Recommend setting each threading "
                "configuration to %d. Set VLLM_SPYRE_UPDATE_THREAD_CONFIG=1 "
                "to do this automatically.", thread_warning, detection_message,
                worker_count, cpus_per_worker)

    def get_max_output_tokens(self, prompt_len: int) -> int:
        """Return the size of biggest ```new_tokens``` of the \
            warmup shapes that fits the prompt length"""
        if self._warmup_shapes is None:
            # ceil division to pad to next block boundary
            padded_prompt_len = math.ceil(
                prompt_len / self._block_size) * self._block_size
            max_new_tokens = (self._config.scheduler_config.max_model_len -
                              padded_prompt_len)
            return max_new_tokens

        max_new_tokens = 1
        for shape in self._warmup_shapes:
            if prompt_len <= shape['prompt_length']:
                max_new_tokens = max(max_new_tokens, shape['new_tokens'])

        return max_new_tokens

    @classmethod
    def configure_granite_3_8b(cls, vllm_config: VllmConfig):
        """
        Configure hard coded values for the model
        https://huggingface.co/ibm-granite/granite-3.3-8b-instruct
        """
        parallel_config = vllm_config.parallel_config

        if parallel_config.world_size != 4:
            # only override configs for TP=4
            return

        tkv_128k = 128 * 1024
        if not os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT"):
            os.environ["VLLM_DT_MAX_BATCH_TKV_LIMIT"] = str(tkv_128k)
            logger.info("Model granite-3.3-8b-instruct and tensor parallel " \
            "size 4 detected. Using VLLM_DT_MAX_BATCH_TKV_LIMIT = %d",
            tkv_128k)
        elif os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT") != str(tkv_128k):
            logger.warning(
                "VLLM_DT_MAX_BATCH_TKV_LIMIT was set to %s, not "
                "overriding to the granite-3.3-8b-instruct default of %d",
                os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT"), tkv_128k)

        # If no HDMA p2psize override was specified, set 256MB
        p2psize_256m = 256 * 1024 * 1024
        if not os.getenv("FLEX_HDMA_P2PSIZE"):
            os.environ["FLEX_HDMA_P2PSIZE"] = str(p2psize_256m)
            logger.info(
                "Model granite-3.3-8b-instruct and tensor parallel size 4 "
                "detected. Using FLEX_HDMA_P2PSIZE = %d", p2psize_256m)
        elif os.getenv("FLEX_HDMA_P2PSIZE") != str(p2psize_256m):
            logger.warning(
                "FLEX_HDMA_P2PSIZE was set to %s, not using the "
                "granite-3.3-8b-instruct default of %d",
                os.getenv("FLEX_HDMA_P2PSIZE"), p2psize_256m)

        # Override the total number of KV cache blocks based on what we know
        # will fit. (Unless user already set `--num-gpu-blocks-override`)
        # TODO: remove this once we have correct free memory info available
        blocks_override = 2080
        if vllm_config.cache_config.num_gpu_blocks_override is None:
            vllm_config.cache_config.num_gpu_blocks_override = blocks_override
            logger.info(
                "Model granite-3.3-8b-instruct and tensor parallel size 4 "
                "detected. Overriding available KV Cache blocks to %d",
                blocks_override)
        elif (vllm_config.cache_config.num_gpu_blocks_override
              != blocks_override):
            logger.warning(
                "--num-gpu-blocks-override was set to %d, not using the "
                "granite-3.3-8b-instruct default of %d",
                vllm_config.cache_config.num_gpu_blocks_override,
                blocks_override)

    @classmethod
    def is_granite_3_8b(cls, model_config: ModelConfig):
        """Returns true if we have a model that looks like
        ibm-granite/granite-3.3-8b-instruct"""
        if not isinstance(model_config.hf_config, GraniteConfig):
            # Not granite at all
            return False

        return (model_config.hf_config.num_hidden_layers == 40
                and model_config.hf_config.max_position_embeddings == 131072
                and model_config.hf_config.hidden_size == 4096
                and model_config.hf_config.vocab_size == 49159
                and model_config.hf_config.num_key_value_heads == 8
                and model_config.hf_config.num_attention_heads == 32)



from .stats_logger import (FileStatLogger, file_stat_logger_factory,
                           patch_async_llm_stat_loggers)

__all__ = [
    "patch_async_llm_stat_loggers", "file_stat_logger_factory",
    "FileStatLogger"
]



"""Verification of continuous batching

Run `python -m pytest tests/e2e/test_spyre_cb.py`.
"""

import pickle
from pathlib import Path
from typing import Any

import pytest
from llm_cache_util import force_engine_shutdown
from openai import BadRequestError
from output_util import (check_output_against_hf, extract_output,
                         generate_spyre_vllm_output)
from spyre_util import (ModelInfo, RemoteOpenAIServer, create_seq_prompt,
                        get_chicken_soup_prompts, skip_unsupported_tp_size)
from vllm import LLM, SamplingParams


@pytest.mark.cb
@pytest.mark.parametrize(
    "backend", [pytest.param("eager", marks=pytest.mark.cpu, id="eager")])
def test_cb_max_tokens(model: ModelInfo, backend: str, max_model_len: int,
                       max_num_seqs: int, monkeypatch: pytest.MonkeyPatch,
                       use_llm_cache):
    """Test that continuous batches of requests that
    are longer than the `max_model_len` are correctly rejected"""
    max_tokens = 20

    overflow_prompt = " ".join(["a"] * max_model_len)

    vllm_sampling_params = SamplingParams(max_tokens=max_tokens,
                                          temperature=0,
                                          ignore_eos=True,
                                          logprobs=0)

    with pytest.raises(ValueError, match="max model context length"):
        generate_spyre_vllm_output(model=model,
                                   prompts=overflow_prompt,
                                   max_model_len=max_model_len,
                                   sampling_params=vllm_sampling_params,
                                   tensor_parallel_size=1,
                                   backend=backend,
                                   max_num_seqs=max_num_seqs,
                                   use_cb=True,
                                   monkeypatch=monkeypatch)


@pytest.mark.cb
@pytest.mark.parametrize("cb", [True])
@pytest.mark.parametrize(
    "backend", [pytest.param("eager", marks=pytest.mark.cpu, id="eager")])
def test_api_cb_rejects_oversized_request(
    remote_openai_server: RemoteOpenAIServer,
    model: ModelInfo,
    backend: str,
    cb: bool,
    max_model_len: int,
    max_num_seqs: int,
):
    """Verify API rejects request that exceed max_model_len with CB enabled"""

    client = remote_openai_server.get_client()
    overflow_prompt = " ".join(["hi"] * max_model_len)
    max_tokens = 10

    with pytest.raises(BadRequestError, match="maximum context length is"):
        client.completions.create(
            model=model.name,
            prompt=overflow_prompt,
            max_tokens=max_tokens,
        )


@pytest.mark.cb
@pytest.mark.parametrize("cb", [True])
@pytest.mark.parametrize(
    "backend", [pytest.param("eager", marks=pytest.mark.cpu, id="eager")])
def test_api_cb_generates_correct_max_tokens(
    remote_openai_server: RemoteOpenAIServer,
    model: ModelInfo,
    backend: str,
    cb: bool,
    max_model_len: int,
    max_num_seqs: int,
):
    """Verify API generates the correct numbers of tokens with CB enabled"""

    client = remote_openai_server.get_client()
    max_tokens = 10

    response = client.completions.create(model=model.name,
                                         prompt=get_chicken_soup_prompts(1),
                                         max_tokens=max_tokens,
                                         temperature=0)

    assert response.usage.completion_tokens == max_tokens


@pytest.mark.compiler_support_32k
@pytest.mark.cb
@pytest.mark.parametrize(
    "backend", [pytest.param("sendnn", marks=pytest.mark.spyre, id="sendnn")])
@pytest.mark.parametrize(
    "tp_size",
    [
        pytest.param(4, marks=pytest.mark.multi),
    ],
    ids=lambda val: f"TP({val})",
)
def test_long_context_batches(
    model: ModelInfo,
    backend: str,
    tp_size: int,
    monkeypatch: pytest.MonkeyPatch,
):
    """Tests continuous batching with various batch sizes and prompt lengths."""

    skip_unsupported_tp_size(tp_size, backend)

    monkeypatch.setenv("VLLM_SPYRE_USE_CB", "1")
    monkeypatch.setenv("VLLM_SPYRE_DYNAMO_BACKEND", backend)
    monkeypatch.setenv("VLLM_SPYRE_OVERRIDE_SIGNALS_HANDLER", "1")

    max_model_len = 32768
    max_num_seqs = 32
    max_tokens = 10

    # (batch_size, prompt_length) pairs
    batch_token_pairs = [
        (32, 512),
        (16, 1500),
        (8, 3000),
        (4, 5000),
        (2, 9000),
        (1, 17000),
    ]

    vllm_model = LLM(model=model.name,
                     tokenizer=model.name,
                     max_model_len=max_model_len,
                     max_num_seqs=max_num_seqs,
                     tensor_parallel_size=tp_size,
                     revision=model.revision)

    sampling_params = SamplingParams(
        max_tokens=max_tokens,
        temperature=0,
        ignore_eos=True,
        logprobs=0,
    )

    for batch_size, token_len in batch_token_pairs:
        prompt = create_seq_prompt(model, token_length=token_len)
        prompts = [prompt] * batch_size

        vllm_outputs = vllm_model.generate(prompts, sampling_params)

        results = []
        for req_output in vllm_outputs:
            result = extract_output(req_output)
            results.append(result)

    check_output_against_hf(
        model=model,
        backend=backend,
        max_new_tokens=max_tokens,
        vllm_results=results,
        prompts=prompts,
    )

    force_engine_shutdown(vllm_model)


@pytest.mark.compiler_support_32k
@pytest.mark.spyre
@pytest.mark.cb
@pytest.mark.parametrize(
    "tp_size",
    [
        pytest.param(4, marks=pytest.mark.multi),
    ],
    ids=lambda val: f"TP({val})",
)
def test_swap_decode_programs_for_cb(
    tp_size: int,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    '''
    
    Validate the runtime's ability to swap between different compiled decode 
    programs for varying batch sizes and TKV.

    The test case consists of 32 small input prompts with specifically chosen 
    max_new_tokens values to trigger different decode programs at runtime.

    The test case structure is as follows:

    - 16 prompts with max_new_tokens @ 1k
    -  8 prompts with max_new_tokens @ 2k
    -  4 prompts with max_new_tokens @ 4k
    -  2 prompts with max_new_tokens @ 8k
    -  1 prompt  with max_new_tokens @ 16k
    -  1 prompt  with max_new_tokens @ 32k
    
    '''

    model = 'ibm-granite/granite-3.3-8b-instruct'
    backend = 'sendnn'
    max_num_seqs = 32

    max_model_len = 32 * 1024  # 32K

    skip_unsupported_tp_size(tp_size, backend)
    prompts = get_chicken_soup_prompts(max_num_seqs)

    create_sampling_params = lambda max_new_tokens: SamplingParams(
        # The prompt will pad to 64 tokens, therefore to match
        # max_model_len/max_new_tokens, we need to decrease by the prompt
        # length
        max_tokens=max_new_tokens - 64,
        temperature=0,
        logprobs=0,  # return logprobs of generated tokens only
        ignore_eos=True)

    p1k = 1 * 1024
    p2k = 2 * 1024
    p4k = 4 * 1024
    p8k = 8 * 1024
    p16k = 16 * 1024
    p32k = 32 * 1024

    sampling_params_1k = [create_sampling_params(p1k) for _ in range(16)]
    sampling_params_2k = [create_sampling_params(p2k) for _ in range(8)]
    sampling_params_4k = [create_sampling_params(p4k) for _ in range(4)]
    sampling_params_8k = [create_sampling_params(p8k) for _ in range(2)]
    sampling_params_16k = [create_sampling_params(p16k) for _ in range(1)]
    sampling_params_32k = [create_sampling_params(p32k) for _ in range(1)]

    sampling_params = sampling_params_1k + sampling_params_2k + \
        sampling_params_4k + sampling_params_8k + sampling_params_16k + \
            sampling_params_32k

    # Read the cache and check beforehand if the cache was written with the
    # expected prompt. We use the filepath of this script to resolve
    # the cache filepaths
    script_directory = Path(__file__).parent.absolute() / 'cache'
    with open(script_directory / 'prompts_8k_bs2.pickle', 'rb') as f:
        cache_result_8k_bs2: list[dict[str, Any]] = pickle.loads(f.read())

    assert cache_result_8k_bs2[0]['prompt'] == prompts[28]
    assert cache_result_8k_bs2[1]['prompt'] == prompts[29]

    with open(script_directory / 'prompts_16k_bs1.pickle', 'rb') as f:
        cache_result_16k_bs1: list[dict[str, Any]] = pickle.loads(f.read())

    assert cache_result_16k_bs1[0]['prompt'] == prompts[30]

    # Generate results from vLLM
    vllm_results = generate_spyre_vllm_output(model=model,
                                              prompts=prompts,
                                              sampling_params=sampling_params,
                                              tensor_parallel_size=tp_size,
                                              backend=backend,
                                              max_num_seqs=max_num_seqs,
                                              monkeypatch=monkeypatch,
                                              max_model_len=max_model_len,
                                              use_cb=True)

    # TODO: dummy validation, currently the outputs do not match with
    # HF cache.

    assert vllm_results is not None
    # Check first from cache, to save computation
    # # 2 @ 8K
    # compare_results(
    #     model=model,
    #     tensor_parallel_size=tp_size,
    #     backend=backend,
    #     vllm_results=vllm_results[28:30],
    #     hf_results=cache_result_8k_bs2,
    #     prompts=prompts[28:30]
    # )

    # # 1 @ 16K
    # compare_results(
    #     model=model,
    #     tensor_parallel_size=tp_size,
    #     backend=backend,
    #     vllm_results=vllm_results[30:31],
    #     hf_results=cache_result_16k_bs1,
    #     prompts=prompts[30:31]
    # )

    # # 16 @ 1K
    # check_output_against_hf(model, backend, p1k, vllm_results[0:16],
    #                         prompts[0:16])

    # # 8 @ 2K
    # check_output_against_hf(model, backend, p2k, vllm_results[16:24],
    #                         prompts[16:24])

    # # 4 @ 4K
    # check_output_against_hf(model, backend, p4k, vllm_results[24:28],
    #                         prompts[24:28])



import copy
import os
from collections import defaultdict, deque
from typing import Any

import pytest
from llm_cache import get_cached_engine
from output_util import (ISCLOSE_ABS_TOL, ISCLOSE_ABS_TOL_QUANTIZATION,
                         compare_results, generate_hf_output)
from spyre_util import ModelInfo, create_random_request
from vllm import SamplingParams
from vllm.transformers_utils.tokenizer import get_tokenizer
from vllm.v1.engine import EngineCoreRequest
from vllm.v1.engine.core import EngineCore

from vllm_spyre.v1.core.scheduler import ContinuousBatchingSpyreScheduler

DISABLE_ASSERTS = False  # used for debugging


def augment_checked_steps(
        checked_steps: list[dict[str, Any]]) -> deque[dict[str, Any]]:
    # Augment checked_steps: add in-between normal decode steps
    checked_steps = deque(checked_steps)
    all_checked_steps = deque()
    prev_step = None
    for step in range(checked_steps[-1]["step"] + 1):
        if checked_steps and step == checked_steps[0]["step"]:
            prev_step = checked_steps.popleft()
            all_checked_steps.append(prev_step)
        elif prev_step is not None:
            assert prev_step["step"] == step - 1
            new_step = copy.deepcopy(prev_step)
            new_step["step"] = step
            new_step["tkv"] += 1
            all_checked_steps.append(new_step)
            prev_step = new_step
    return all_checked_steps


def generate_prompts(
    model: ModelInfo,
    steps_add_reqs: list[int],
    seqs_max_tokens: list[int],
    prompts_lengths: list[int],
):
    generated_prompts = []

    # Create random requests of specified lengths and max_tokens
    # Need to do before setting up the vLLM engine, otherwise test random seed
    # will be overridden
    sorted_reqs_params = zip(steps_add_reqs, seqs_max_tokens, prompts_lengths)
    requests: deque[tuple[int, EngineCoreRequest]] = deque()
    for i, (add_step, max_tokens,
            prompt_length) in enumerate(sorted_reqs_params):
        # ignoring eos because we want to force the decoding to finish
        # after max_tokens exactly
        sampling_params = SamplingParams(max_tokens=max_tokens,
                                         temperature=0.0,
                                         logprobs=0,
                                         ignore_eos=True)
        request = create_random_request(request_id=i,
                                        num_tokens=prompt_length,
                                        sampling_params=sampling_params,
                                        model=model)
        requests.append((add_step, request))
        # NOTE: It is going to be decoded later
        generated_prompts.append(request.prompt_token_ids)

    return generated_prompts, requests


def check_scheduler_inference_steps(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    seqs_max_tokens: list[int],
    prompts_lengths: list[int],
    steps_add_reqs: list[int],
    checked_steps: list[dict[str, Any]],
    max_num_seqs: int,
    max_model_len: int,
    available_blocks: int,
    max_batch_tkv_limit: int = -1,
    use_cb: bool = True,
):
    """
    Test the scheduler execution by comparing the scheduler attributes at each 
    step with the provided reference values in 'checked_steps'.
    
    The missing steps from 'checked_steps' are automatically generated as decode
    steps, based on the existing elements in the list. For that to work, all the
    prefill steps and the first decode step after them needs be added to 
    'checked_steps'
    """

    # Input parameters sanity check, not actual testing
    # ------
    if not (len(prompts_lengths) == len(seqs_max_tokens)
            and len(prompts_lengths) == len(steps_add_reqs)):
        raise ValueError(
            "Number of prompts should be consistent with number of max tokens."
        )

    if not (steps_add_reqs == sorted(steps_add_reqs)
            and steps_add_reqs[0] == 0):
        raise ValueError(
            "The list of steps where requests are added should be increasing "
            "start with 0")

    if not (checked_steps == sorted(checked_steps, key=lambda x: x["step"])
            and len(checked_steps) == len(set(x["step"]
                                              for x in checked_steps))):
        raise ValueError(
            "List of checked steps needs to be of increasing order of step")
    # ------

    collected_outputs = defaultdict(lambda: {
        "token_ids": [],
        "logprobs": [],
        "text": "",
        "tokens": []
    })

    prompts, requests = generate_prompts(model, steps_add_reqs,
                                         seqs_max_tokens, prompts_lengths)

    hf_results = generate_hf_output(
        model=model,
        prompts=prompts,
        max_new_tokens=seqs_max_tokens,
        ignore_eos=True,
    )

    abs_tol = ISCLOSE_ABS_TOL_QUANTIZATION if model.is_quantized \
         else ISCLOSE_ABS_TOL
    # inject expectation.
    # json is fine to transfer between vllm subprocesses using pickle
    for idx, (req, hf) in enumerate(zip(requests, hf_results)):
        req[1].sampling_params.extra_args = {
            "golden_token_injector": {
                "expected_token_ids": hf['token_ids'],
                "expected_logprobs": hf['logprobs'],
                "error_threshold": abs_tol,
                "label": f"#{idx}"
            }
        }

    # Setup the engine
    engine_core: EngineCore = get_cached_engine(
        model=model,
        max_model_len=max_model_len,
        max_num_seqs=max_num_seqs,
        available_blocks=available_blocks,
        backend=backend,
        monkeypatch=monkeypatch)
    scheduler: ContinuousBatchingSpyreScheduler = engine_core.scheduler

    tokenizer = get_tokenizer(model.name, revision=model.revision)
    # clear the cache of function scheduler.check_batch_tkv_limit()
    scheduler._cache_check_batch_tkv_limit.clear()

    # Override the TKV limit in the scheduler if needed
    if max_batch_tkv_limit >= 0:
        scheduler.max_batch_tkv_limit = max_batch_tkv_limit
    else:
        # This default value is set by platform.py
        scheduler.max_batch_tkv_limit = int(
            os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT"))

    # In-between steps are added as normal decode steps
    checked_steps = augment_checked_steps(checked_steps)

    # Run steps, until last step from 'checked_steps' is reached
    request_outputs = []
    requested_blocks, reserved_blocks = {}, {}
    for step in range(checked_steps[-1]['step'] + 1):
        # Add requests for this step
        while requests and requests[0][0] == step:
            engine_core.add_request(requests.popleft()[1])

        # Check step if it is in the provided list of steps to check
        if checked_steps and step == checked_steps[0]["step"]:
            step_ref = checked_steps.popleft()

            waiting = [r.request_id for r in scheduler.waiting]
            running = [r.request_id for r in scheduler.running]
            out_reqs_ids = [r.request_id for r in request_outputs]
            out_reqs_finished = [
                r.request_id for r in request_outputs if r.finished
            ]

            assert DISABLE_ASSERTS or (scheduler.tkv == step_ref["tkv"]
                                       ), f"Step {step}, tkv: {scheduler.tkv}"
            assert (DISABLE_ASSERTS or waiting
                    == step_ref["waiting"]), f"Step {step}, waiting: {waiting}"
            assert (DISABLE_ASSERTS or running
                    == step_ref["running"]), f"Step {step}, running: {running}"
            assert DISABLE_ASSERTS or (
                out_reqs_ids == step_ref["request_outputs"]
            ), f"Step {step}, request outputs: {out_reqs_ids}"

            ref_finished_reqs = step_ref.get("finished_requests", [])
            assert DISABLE_ASSERTS or (
                out_reqs_finished == ref_finished_reqs
            ), f"Step {step}, finished request output: {out_reqs_finished}"

            # checking the scheduler handling of free and reserved blocks
            n_blocks = (engine_core.model_executor.driver_worker.worker.
                        model_runner.n_blocks)
            n_reserved_blocks = n_blocks - scheduler.n_free_blocks
            req_ids2blocks = (engine_core.model_executor.driver_worker.worker.
                              model_runner.req_ids2blocks)
            req_ids2reserved_blocks = (
                engine_core.model_executor.driver_worker.worker.model_runner.
                req_ids2reserved_blocks)
            n_used_blocks = sum(
                [len(blocks) for blocks in req_ids2blocks.values()])

            if step > 0:
                assert DISABLE_ASSERTS or (
                    n_reserved_blocks == step_ref["n_reserved_blocks"]
                ), f"Step {step}, n_reserved_blocks: {n_reserved_blocks}"
                assert DISABLE_ASSERTS or (
                    n_used_blocks == step_ref["n_used_blocks"]
                ), f"Step {step}, n_used_blocks: {n_used_blocks}"

            assert DISABLE_ASSERTS or len(req_ids2blocks) == len(
                req_ids2reserved_blocks)
            for req_id in req_ids2blocks:
                # current number of used blocks should be less than reserved
                assert (DISABLE_ASSERTS or len(req_ids2blocks[req_id])
                        <= req_ids2reserved_blocks[req_id])
                # update requested/reserved blocks to check in last step
                # Note: overwrite and not max
                # because of reduce_left_padding()
                requested_blocks[req_id] = len(req_ids2blocks[req_id])
                reserved_blocks[req_id] = req_ids2reserved_blocks[req_id]

        # last step: check that sequences used all their reserved blocks
        # Note: no early stopping, all sequences produce max_num_tokens
        if len(checked_steps) == 0:
            for req_id in requested_blocks:
                assert (DISABLE_ASSERTS
                        or requested_blocks[req_id] == reserved_blocks[req_id])

        # Perform next step
        step_output = engine_core.step()
        engine_core_output = step_output[0].get(0)
        request_outputs = (engine_core_output.outputs
                           if engine_core_output is not None else [])

        for output in request_outputs:
            new_token_ids = output.new_token_ids
            new_logprobs = output.new_logprobs.logprobs
            assert DISABLE_ASSERTS or len(new_token_ids) == 1 and len(
                new_logprobs) == 1

            collected_outputs[output.request_id]["token_ids"].append(
                new_token_ids[0])
            collected_outputs[output.request_id]["logprobs"].append(
                new_logprobs[0][0])
            collected_outputs[output.request_id]["tokens"].append(
                tokenizer.decode(new_token_ids[0]))

    for k in collected_outputs:
        collected_outputs[k]['text'] = tokenizer.decode(
            collected_outputs[k]['token_ids'])
    output_keys = sorted(int(k) for k in collected_outputs)
    assert (DISABLE_ASSERTS
            or output_keys[0] == 0 and output_keys[-1] == len(output_keys) - 1)

    # convert dict of dicts to ordered list and make values immutable
    vllm_results = []
    for k in output_keys:
        output = collected_outputs[str(k)]
        for k, list_values in output.items():
            if isinstance(list_values, list):
                output[k] = tuple(list_values)
        vllm_results.append(output)

    compare_results(model=model,
                    tensor_parallel_size=1,
                    backend=backend,
                    vllm_results=vllm_results,
                    hf_results=hf_results,
                    prompts=prompts)



"""Verification of vLLM output by comparing with HF
with VLLM_SPYRE_MAX_LOAD_PROCESSES enabled.

Run `python -m pytest tests/e2e/test_stagger_spyre_basic.py`.
"""

import pytest
from output_util import validate_vllm_vs_hf_output
from spyre_util import (ModelInfo, get_chicken_soup_prompts,
                        skip_unsupported_tp_size)
from vllm import SamplingParams


def test_stagger_output(model: ModelInfo, tp_size: int, backend: str, cb: int,
                        max_num_seqs: int, max_model_len: int, warmup_shapes,
                        monkeypatch: pytest.MonkeyPatch,
                        use_llm_cache) -> None:
    '''
    This test verifies that generated output is still correct
    when stagger mode is enabled.
    VLLM_SPYRE_MAX_LOAD_PROCESSES is set to 1, allowing
    only a single worker to load or compile the model at
    a time.
    '''

    skip_unsupported_tp_size(tp_size, backend)
    monkeypatch.setenv("VLLM_SPYRE_MAX_LOAD_PROCESSES", "1")

    prompts = get_chicken_soup_prompts(4)
    warmup_shape = (64, 20, 4)

    kwargs = ({
        "max_num_seqs": max_num_seqs,
        "use_cb": True,
    } if cb == 1 else {
        "warmup_shapes": warmup_shapes
    })

    max_new_tokens = warmup_shape[1]

    vllm_sampling_params = SamplingParams(
        max_tokens=max_new_tokens,
        temperature=0,
        logprobs=0,  # return logprobs of generated tokens only
        ignore_eos=True)

    validate_vllm_vs_hf_output(model=model,
                               prompts=prompts,
                               sampling_params=vllm_sampling_params,
                               tensor_parallel_size=tp_size,
                               backend=backend,
                               monkeypatch=monkeypatch,
                               max_model_len=max_model_len,
                               max_new_tokens=max_new_tokens,
                               **kwargs)



import json
import random
from typing import Optional

import pytest
import torch
from spyre_util import ModelInfo
from vllm.sampling_params import SamplingParams
from vllm.v1.sample.logits_processor import BatchUpdateBuilder, LogitsProcessor

from vllm_spyre.v1.sample.golden_token_injector import GoldenTokenInjector


class DummyVllmConfig:

    def __init__(self, model_config: ModelInfo):
        self.model_config = DummyModelConfig(model_config.name,
                                             model_config.revision)


class DummyModelConfig:

    def __init__(self, tokenizer: str, revision: Optional[str]):
        self.tokenizer = tokenizer
        self.revision = revision
        self.tokenizer_revision = revision
        self.tokenizer_mode = None
        self.trust_remote_code = True


def step(batch_update_builder: BatchUpdateBuilder, lp: LogitsProcessor,
         logits: torch.Tensor, batch_output_tokens: list[list[int]]):

    assert logits.shape[0] == len(batch_output_tokens)

    # This is called at each execute model in spyre model runner update_states
    lp.update_state(None)

    out_logits = lp.apply(logits.clone())

    for idx, output_tokens in enumerate(batch_output_tokens):
        # just append a random token
        token_idx = out_logits[idx].argmax(dim=-1).reshape(-1).item()
        output_tokens.append(token_idx)

    return out_logits


def generate_logits(vocab_size: int, batch_size: int = 1):

    return torch.tensor(list(range(vocab_size)) * batch_size,
                        dtype=torch.float32).reshape((batch_size, vocab_size))


@pytest.mark.cpu
@pytest.mark.parametrize("arg_as_string", [True, False])
def test_gti_basic_correctness(model: ModelInfo, arg_as_string: bool):

    device = torch.device('cpu')
    gti = GoldenTokenInjector(DummyVllmConfig(model), device, False)

    vocab_size = gti.tokenizer.vocab_size

    batch_update_builder = BatchUpdateBuilder()
    batch_output_tokens = []
    logits = generate_logits(vocab_size, 1)

    expected_tokens_count = 8

    expected_token_ids = [
        random.randint(0, vocab_size) for _ in range(expected_tokens_count)
    ]
    gti_args = {
        "expected_token_ids": \
            expected_token_ids,
    }

    if arg_as_string:
        gti_args = json.dumps(gti_args)

    params = SamplingParams(extra_args={"golden_token_injector": gti_args})

    prompt_tokens = [random.randint(0, vocab_size) for _ in range(8)]
    batch_update_builder.added.append(
        (0, params, prompt_tokens, batch_output_tokens))
    batch_update = batch_update_builder.get_and_reset(1)
    gti.update_state(batch_update)

    for current_idx in range(expected_tokens_count):
        logits = generate_logits(vocab_size, 1)
        step(batch_update_builder, gti, logits, [batch_output_tokens])
        assert batch_output_tokens[current_idx] == expected_token_ids[
            current_idx]


@pytest.mark.cpu
def test_gti_out_of_range_expected_tokens(model: ModelInfo):
    # TODO: this test is a huge copy paste from the above
    # improve it later to better reuse
    device = torch.device('cpu')
    gti = GoldenTokenInjector(DummyVllmConfig(model), device, False)

    vocab_size = gti.tokenizer.vocab_size

    batch_update_builder = BatchUpdateBuilder()
    batch_output_tokens = []
    logits = generate_logits(vocab_size, 1)

    expected_tokens_count = 8

    expected_token_ids = [
        random.randint(0, vocab_size) for _ in range(expected_tokens_count)
    ]
    gti_args = {
        "expected_token_ids": \
            expected_token_ids,
    }

    params = SamplingParams(extra_args={"golden_token_injector": gti_args})

    prompt_tokens = [random.randint(0, vocab_size) for _ in range(8)]
    batch_update_builder.added.append(
        (0, params, prompt_tokens, batch_output_tokens))
    batch_update = batch_update_builder.get_and_reset(1)
    gti.update_state(batch_update)

    # Inject correctly
    for current_idx in range(expected_tokens_count):
        logits = generate_logits(vocab_size, 1)
        step(batch_update_builder, gti, logits, [batch_output_tokens])
        assert batch_output_tokens[current_idx] == expected_token_ids[
            current_idx]

    # Cannot inject anymore
    logits = generate_logits(vocab_size, 1)
    out_logits = step(batch_update_builder, gti, logits, [batch_output_tokens])
    # Keep logits same
    assert torch.allclose(logits, out_logits)
